﻿# Extract Metadata from a single VMDK-File, or all VMDK-Files in a Directory
[CmdletBinding()]
Param(
    [Parameter(Position=0)]
    [String] $FilePath
)

# ---- Extract Metadata from VMDK file ---
function Get-VmdkMetadata() {
  [CmdletBinding()]
  Param(
    [Parameter(Mandatory=$True)]
    [String] $FileName
  )

  $HeaderSize = 1536
  [byte[]] $MagicVMDK = ('VMDK').ToCharArray(); [array]::Reverse($MagicVMDK)

  if (Test-Path -Path $FileName -PathType Leaf) {
    $FileSize = (Get-Item $FileName).Length
    Write-MyLog ("+++ Processing File [{0}] (Size: {1:N0} Bytes) ..." -f $FileName, $FileSize) -Logfile $myLogFile
    if ($FileSize -ge $HeaderSize) {
        $myOutFile = "{0}\Metadata-{1}-{2}.bin" -f $myPath, (Get-Item $FileName).BaseName, (Get-Item $FileName).Length
        $VmdkHeader = New-Object byte[] $HeaderSize
        $fs  = New-Object IO.FileStream $FileName, 'Open', 'Read'
        $reader = New-Object IO.BinaryReader($fs)
        $BytesRead = $reader.Read($VmdkHeader, 0, $HeaderSize)
        if ((compare $VmdkHeader[0..3] $MagicVMDK).Count -gt 0) {
            Write-MyLog "Invalid Magic Number detected." -Logfile $myLogFile -Level ERROR
            [System.IO.File]::WriteAllBytes($myOutFile, $VmdkHeader)
        } else {
            $MetadataSize = 0; foreach ($i in $VmdkHeader[71..64]) {$MetadataSize = ($MetadataSize * 256) + $i}
            Write-MyLog ("Number of sectors occupied by the metadata = {0:N0} (Size: {1:N0} Bytes)" -f $MetadataSize, ($MetadataSize * 512))  -Logfile $myLogFile
            $MetadataSize *= 512
            if ($FileSize -lt $MetadataSize) {Write-MyLog "File size too small. Possibly truncated." -Logfile $myLogFile -Level FATAL}
            $VmdkMetadata = New-Object byte[] $MetadataSize
            $reader.BaseStream.Position = 0
            $BytesRead = $reader.Read($VmdkMetadata, 0, $MetadataSize)
            [System.IO.File]::WriteAllBytes($myOutFile, $VmdkMetadata)
        }
        $fs.Close()
    } else {
        Write-MyLog "File size too small."  -Logfile $myLogFile -Level ERROR
    }
  } else {
    Write-MyLog ("File [{0}] does not exist." -f $FileName)  -Logfile $myLogFile -Level ERROR
  }
}

# -- Log-Output ---
Function Write-MyLog {
    [CmdletBinding()]
    Param(
    [Parameter(Mandatory=$True, Position=0)]
    [string] $Message,
    [Parameter(Mandatory=$False)]
    [ValidateSet("INFO","WARN","ERROR","FATAL")]
    [String] $Level = "INFO",
    [Parameter(Mandatory=$False)]
    [string] $Logfile
    )
    $MyLine = "{0}: {1}" -f $Level.PadRight(6, " "), $Message
    Write-Output $MyLine
    If($Logfile) {Add-Content $Logfile -Value $MyLine}
}

# --- Main ---
$myPath    = $PSScriptRoot
$myLogFile = $mypath + "\Metadata-" + (Get-Item $MyInvocation.MyCommand.Path).BaseName + ".log"

if (!$FilePath) {$FilePath = $myPath}
try {
    if ((Get-Item $FilePath -ErrorAction Stop) -is [System.IO.DirectoryInfo]) {
        Get-ChildItem $FilePath -Filter *.vmdk | ForEach-Object {Get-VmdkMetadata $_.FullName}
    } else {
        Get-VmdkMetadata -FileName $FilePath
    }
}
catch {
    Write-MyLog ("File or folder {0} does not exist!" -f $FilePath) -Logfile $myLogFile -Level ERROR
}

